package palm;

// Sort of based on the Waba routines that simulated a PDB file so you could
// test your Palm code on the PC.. but that never really created a file.
// Won't say that this is pretty... but it works for now so I could read
// a particular file.
// Nick Vrtis - nick@vrtisworks.com

import java.io.IOException;
import java.io.BufferedInputStream;
import java.net.URL;
import java.net.URLConnection;


/**
 * A PDB file is the version of a Palm file that is stored on
 * the PC by the synchronization program.  This class is designed to allow
 * a PC based Java program to access these files.
 * <p>
 * This is an incomplete version that I put together in order to read a
 * particular file.  Eventually, I'll replace it with the more complete
 * version.
 */

public class PDBFile extends Object
{
/** Read-only open mode. */
public static final int READ_ONLY  = 1;
/** Write-only open mode. */
public static final int WRITE_ONLY = 2;
/** Read-write open mode. */
public static final int READ_WRITE = READ_ONLY+WRITE_ONLY
; // READ | WRITE
/** Create open mode. Used to create a database if one does not exist. */
public static final int CREATE = 4+READ_WRITE;

private boolean isOpen=false;
private int openMode;
private byte [] theFile;						//This is the whole input file
private int [][] theRecordList;	//An array of ints for each record
private int recordPos;					//Current record number
private byte [] currentData = new byte[1];		//Copy of data for the current record
private int [] currentInfo;			//Copy of 'theRecordList' for the current record
private int cursor;							//Position within the current record.

/**
 * Opens a Palm file with the given name and mode. If mode is CREATE, the
 * file will be created if it does not exist.
 * <p>
 * For PalmOS: A PalmOS creator id and type can be specified by appending
 * a 4 character creator id and 4 character type to the name seperated
 * by periods. For example:
 * <pre>
 * Catalog c = new Catalog("MyCatalog.CRTR.TYPE", Catalog.CREATE);
 * </pre>
 * Will create a PalmOS database with the name "MyCatalog", creator id
 * of "CRTR" and type of "TYPE".
 * <p>
 * If no creator id and type is specified, the creator id will default
 * to the creator id of current waba program and the type will default
 * to "DATA".
 * <p>
 * Under PalmOS, the name of the catalog must be 31 characters or less,
 * not including the creator id and type. Windows CE supports a 32
 * character catalog name but to maintain compatibility with PalmOS,
 * you should use 31 characters maximum for the name of the catalog.
 * @param name PC file name.
 * @param mode one of READ_ONLY, WRITE_ONLY, READ_WRITE or CREATE
 */


public PDBFile (URL url, String name, int mode)
			throws IOException, IllegalArgumentException	{
	String s;
  URL theURL;
  URLConnection uc;
  BufferedInputStream theBIS;
  int theSize;
  int bytesRead;

	if (mode == CREATE)	{
		//		s = "Catalog: creating " + name;
		throw (new IllegalArgumentException("Create not available"));
	}
	else {
  	if (mode==READ_ONLY)
    	s="r";
    else
    	//if (mode==WRITE_ONLY)
      //	s="w";
      //else
      //	if (mode==READ_WRITE)
      //  	s="rw";
      //  else
        	throw (new IllegalArgumentException("Mode is invalid"));

 		theURL = new URL(url.getProtocol(),url.getHost(),url.getPort(),url.getFile()+name);
	 	//System.out.println("H="+theURL.getHost()+" F="+theURL.getFile()+" P="+theURL.getProtocol()+" T="+theURL.getPort());
  	uc = theURL.openConnection();
	  uc.setDoInput(true);
  	uc.connect();
	  theSize = uc.getContentLength();
	  //System.out.println("Bytes="+theSize);
  	theFile = new byte[theSize];
	  theBIS = new BufferedInputStream(theURL.openStream());
  	bytesRead = theBIS.read(theFile,0,theFile.length);
	  //System.out.println("Read in="+bytesRead);
  	theBIS.close();

	  retrieveRecordInfo();			//Go find all the pointers to records
	}
	openMode = mode;
	}

/**
 * Closes the catalog. Returns true if the operation is successful and false
 * otherwise.
 */

public boolean close()
	{
//	System.out.println("Catalog: closing " + palmName);
	if (!isOpen)
		return false;
	isOpen = false;
	recordPos = -1;
	return true;
	}

/**
 * Returns the number of records in the catalog or -1 if the catalog is not open.
 */

public int getRecordCount()	{
	if (!isOpen)
		return -1;
	return theRecordList.length;
	}

/**
 * Returns the size of the current record in bytes or -1 if there is no
 * current record.
 */

public int getRecordSize()
	{
	if (recordPos == -1)
		return -1;
	return currentData.length;
	}

/**
 * Returns true if the catalog is open and false otherwise. This can
 * be used to check if opening or creating a catalog was successful.
 */

public boolean isOpen()
	{
	return isOpen;
	}

/**
 * Sets the current record position and locks the given record. The value
 * -1 can be passed to unset and unlock the current record. If the operation
 * is succesful, true is returned and the read/write cursor is set to
 * the beginning of the record. Otherwise an exception is returned.
 * @param the record within the file to be set as the current record.
 */

public void setRecordPos(int pos) throws IOException, IllegalArgumentException {
	recordPos=-1;		//In case of IOException or IllegalArgument or just a reset.
  if (pos==-1)		//This is OK
  	return;
	if (pos < 0 || pos >= theRecordList.length)	{
  	throw(new IllegalArgumentException("Position ("+pos+") invalid."));
	}
  currentInfo = theRecordList[pos];
  if (currentData.length != currentInfo[3])	//Make sure got enough room already
  	currentData = new byte[currentInfo[3]];	//NO-allocate new space
  System.arraycopy(theFile,currentInfo[0],currentData,0,currentInfo[3]);
	recordPos = pos;
	cursor = 0;
	return;
	}

/**
 * Reads bytes from the current record into a byte array. Returns the
 * number of bytes actually read or -1 if an error prevented the
 * read operation from occurring. After the read is complete, the location of
 * the cursor in the current record (where read and write operations start from)
 * is advanced the number of bytes read.
 * @param buf the byte array to read data into
 * @param start the start position in the array
 * @param count the number of bytes to read
 */

public int readBytes(byte buf[], int start, int count) {
	int y=count;
  if ((count+cursor)>currentData.length)
  	y=currentData.length-cursor;
  if (y==0)
  	return y;									//Nothing to copy
  System.arraycopy(theFile,currentInfo[0]+cursor,buf,start,y);
  cursor+=y;
	return y;
	}

/**
 * Advances the cursor in the current record a number of bytes. The cursor
 * defines where read and write operations start from in the record. Returns
 * the number of bytes actually skipped or -1 if an error occurs.
 * @param count the number of bytes to skip
 */

public int skipBytes(int count)	{
	if (recordPos == -1)	//See if positioned at valid record
		return -1;					//Return if not
	int skipped=count;		//else assume we can skip the whole thing
	if (cursor + skipped > currentInfo[3]) {	//Check for trying to skip past end
  	skipped = currentInfo[3]-cursor;				//YES-only can skip to end
  }
//	System.out.println("Catalog: skipping ahead " + count + " bytes");
	cursor += skipped;	//bump cursor
	return skipped;			//and return what got skipped
	}


/**
	* Reads the record list information to get the initial information about each
  * record in the database.
  */
private void retrieveRecordInfo() {
  	int nextRLE=72;
    int nbrRecs;
    int totalRecs=0;
    int posn;

    while (nextRLE!=0) {					//Chase chain to get total records first
    	posn=nextRLE;
      nextRLE = 0+((theFile[posn]&0xff)<<24)+((theFile[posn+1]&0xff)<<16)
      					+((theFile[posn+2]&0xff)<<8)+(theFile[posn+3]&0xff);
      totalRecs = totalRecs+((theFile[posn+4]&0xff)<<8)+(theFile[posn+5]&0xff);
	   	//System.out.println("nextRLE="+nextRLE+" totalRecs="+totalRecs);
    }
    theRecordList = new int [totalRecs][];	//Allocate array to hold list info
    nextRLE=72;									//Reset back to start of chain
    totalRecs=0;
    while (nextRLE!=0) {
    	posn=nextRLE;			//Save
      nextRLE = 0+((theFile[posn]&0xff)<<24)+((theFile[posn+1]&0xff)<<16)
      					+((theFile[posn+2]&0xff)<<8)+(theFile[posn+3]&0xff);
      nbrRecs = 0+((theFile[posn+4]&0xff)<<8)+(theFile[posn+5]&0xff);
      posn+=6;					//Bump to start of record info
      for (int x=0; x<nbrRecs; x++,totalRecs++,posn+=8) {
      	//System.out.println("Rec#="+x+" position="+posn);
      	int [] ri = new int [4];		//Allocate an array to hold the record info
        ri[0]=0+((theFile[posn]&0xff)<<24)+((theFile[posn+1]&0xff)<<16)
        			+((theFile[posn+2]&0xff)<<8)+(theFile[posn+3]&0xff);
        ri[1]=theFile[posn+4]&0xff;
        ri[2]=0+((theFile[posn+5]&0xff)<<16)+((theFile[posn+6]&0xff)<<8)
        			+(theFile[posn+7]&0xff);
        ri[3]=0;						//Will get filled in later
        theRecordList[totalRecs]=ri;
        if (totalRecs>0) {	//length = difference between addresses
        	theRecordList[totalRecs-1][3]= ri[0]-theRecordList[totalRecs-1][0];
        }
      }
    }
    if (totalRecs>0) {
  	  theRecordList[totalRecs-1][3]=theFile.length-theRecordList[totalRecs-1][0];
    }
    //System.out.println("Record Count="+theRecordList.length);
	}
}