// This is an adaptation of ideas from a number of poetry writing programs I
// found.  Sorry to not give credit, but don't remember where I found them.
// I wanted a program to generate a phrase that I could use the first letter
// of each word as part of a password.  Some people can remember phrases easier
// than a random string of characters.
// Nick Vrtis - nick@vrtisworks.com

import java.awt.TextArea;

public class PasswordPhrase extends Object {

	//The words don't need to be in alpha order, but it helps avoid duplicates
  //NOTE: These need to be such that 'singular' just adds an 's' on the end
  private String [] verbs = {"argue","bite","buy","crash","cut","drink",
  	"drive","drop","eat","enter","exhale","fall","find","follow","get",
    "give","grasp","hit","hurt","ignore","imagine","jump","jab","kick","kneel",
    "lick","lift","look","melt","mold","nag","order","open","overflow",
    "place","plead","quiz","read","run","sing","sip","slam","smell","take",
    "touch","understand","usurp","vex","wave","whisper"};
  private boolean [] verbsUsed = new boolean[verbs.length];
  private String [] articles = {"a","all","each","her","his","many","my","no",
  	"one","some","that","the","their","this","your"};
  private boolean [] articlesUsed = new boolean[articles.length];
  private String [] adjectives = {"ancient","angry","blue","bland","clean",
  	"cold","cool","dirty","dark","effective","evil","filthy","flat","folded",
    "green","greasy","hot","icy","juicy","kind","likable","lovely",
    "moldy","necessary","nice","old","odd","plastic","pretty","quiet","quick",
    "red","round","slippery","smelly","tame","timid","uniform","upright",
    "vibrant","wild","white","yellow"};
  private boolean [] adjectivesUsed = new boolean[adjectives.length];
  //NOTE: These need to be such that plural is just an 's' on the end
  private String [] nouns = {"arm","apple","bear","boat",
  	"book","bottle","car","cat","chicken","clock","computer","cow","dog","door",
    "eagle","elbow","finger","file","ghost","hose","hotel","image","jack","jar",
    "key","kite","lamp","map","mixer","mother","name","napkin","nose",
    "oar","opener","pear","printer","quill","rat","rainbow","shoe",
    "snake","soap","train","truck","thumb","umbrella","uncle","vessel",
    "video","vowel","wrench","window","xray","zebra"};
  private boolean [] nounsUsed = new boolean[nouns.length];
  private String [] adverbs = {"abruptly","always","basically","brightly",
  	"cheerfully","deliberately","eagerly","finally","grimly","happily",
    "hysterically","immediately","immensely","justly","kindly",
    "lightly","loudly","moodily","never","painfully","quietly","quickly",
    "rapidly","rarely","remorselessly","sadly","slowly","suddenly","timidly",
    "uncomfortably","undesirably","wetly","wildly"};
  private boolean [] adverbsUsed = new boolean[adverbs.length];

  private static final String VOWELS="aeiou";
	private int resultSize;
  private TextArea theArea;
  private String [] wordList;
  private char [] charList;
  private int wordPosition;


	public PasswordPhrase(TextArea ta) {
  	theArea = ta;
	}

  public String generate(int sz) {

  	resultSize = sz;						//Save the size of the resulting word
  	if (Math.random()>0.5)
    	return newFormat1Word();
    else
    	return newFormat2Word();
  }

  private String newFormat1Word() {
  	String aWord;
    boolean plural;
    int digitNoun;
    int articlePosition;
    int length;
    int x;
    int [] opts = new int[3];	//0=adj1; 1=adj2; 2=adv

  	//Format 1 is 'Art' 'Adj(0-3)' 'Noun' 'Adv(0-1)' 'Verb' 'Art' 'Adj(0-3)' 'Noun' '.'
    opts[0]=3;					//Start with max
    opts[1]=3;
    opts[2]=1;
    length = 12;				//This is starting length
    while (length>resultSize) {	//Trim total to requested Size
    	x = random(3);			//Pick one to trim
      if (opts[x]>0) {
      	opts[x]--;
        length--;
      }
	    //System.out.println("Adj1="+opts[0]+" Adj2="+opts[1]+" Adv="+opts[2]+" Length="+length);
    }

  	initWordsUsed();
    wordList = new String[length];
    charList = new char[length];

    wordPosition=0;
    //Start with article
    articlePosition = wordPosition;
    aWord = pickAWord(articles,articlesUsed);
   	plural=isPlural(aWord);
    addAWord(aWord);
    //Then maybe some adjectives
		while (opts[0]>0) {
    	addAWord(pickAWord(adjectives,adjectivesUsed));
      opts[0]--;
    }
   	aWord = pickAWord(nouns,nounsUsed);
    if (plural)
    	addAWord(aWord+"s");
    else
    	addAWord(aWord);
    //Adjust if 'A' 'vowel'
    if (wordList[articlePosition].equals("A") &&
    		VOWELS.indexOf(wordList[articlePosition+1].charAt(0))>=0)
    	wordList[articlePosition]="An";
    //Then maybe some adverbs
    while (opts[2]>0) {
	    addAWord(pickAWord(adverbs,adverbsUsed));
      opts[2]--;
    }
    //Then the verb (might need to adjust for plural noun subject
    if (plural)
	    addAWord(pickAWord(verbs,verbsUsed));
    else
    	addAWord(pickAWord(verbs,verbsUsed)+"s");
    //Then another article
    articlePosition = wordPosition;
    aWord = pickAWord(articles,articlesUsed);
   	plural=isPlural(aWord);
    addAWord(aWord);
    //Then maybe some adjectives
    while (opts[1]>0) {
    	addAWord(pickAWord(adjectives,adjectivesUsed));
      opts[1]--;
    }
   	aWord = pickAWord(nouns,nounsUsed);
    if (plural)
    	addAWord(aWord+"s");
    else
    	addAWord(aWord);
    //Adjust if 'a' 'vowel'
    if (wordList[articlePosition].equals("a") &&
    		VOWELS.indexOf(wordList[articlePosition+1].charAt(0))>=0)
    	wordList[articlePosition]="an";
    return finishPanel(".");
  }

  private String newFormat2Word() {

  	if (resultSize>7) {				//Max size for this format is 7
    	return newFormat1Word();			//Use format 1 if min is too big
    }

	  //Format 2 is 'Art' 'Adj(0-3)' 'Noun' 'Adv(0-1)' 'Verb' '!'
  	String aWord;
    boolean plural;
    int articlePosition;
    int verbPosition;
    int [] opts = new int[2];	//0=adj1; 2=adv
    int x;
    int length;

    opts[0]=3;								//Start with max
    opts[1]=1;
    length =7;	//See what total length is
    while (length>resultSize) {	//Trim total to requested size
    	x = random(2);			//Pick one to trim
      if (opts[x]>0) {
      	opts[x]--;
        length--;
      }
	    //System.out.println("Adj1="+opts[0]+" Adv="+opts[1]+" Length="+length);
    }

  	initWordsUsed();
    wordList = new String[length];
    charList = new char[length];

    wordPosition=0;
    //An article
    articlePosition = wordPosition;
    aWord = pickAWord(articles,articlesUsed);
   	plural=isPlural(aWord);
    addAWord(aWord);
    //Then maybe some adjectives
		while (opts[0]>0) {
    	addAWord(pickAWord(adjectives,adjectivesUsed));
      opts[0]--;
    }
    //Then a noun (might be plural or a digit)
   	aWord = pickAWord(nouns,nounsUsed);
    if (plural)
    	addAWord(aWord+"s");
    else
    	addAWord(aWord);
    //Adjust if 'A' 'vowel'
    if (wordList[articlePosition].equals("a") &&
    		VOWELS.indexOf(wordList[articlePosition+1].charAt(0))>=0)
    	wordList[articlePosition]="an";
    //Maybe some adverbs
    while (opts[1]>0) {
	    addAWord(pickAWord(adverbs,adverbsUsed));
      opts[1]--;
    }
    //Then the verb
    verbPosition = wordPosition;				//Might need to adjust for plurals
    if(plural)
    	addAWord(pickAWord(verbs,verbsUsed));
    else
	   	addAWord(pickAWord(verbs,verbsUsed)+"s");
    return finishPanel("!");
  }

  private int random(int r) {
  	return (int)(Math.random()*r);
  }

  private void addAWord(String word) {

  	charList[wordPosition]=word.toUpperCase().charAt(0);
  	if (wordPosition==0)
    	wordList[wordPosition]= word.substring(0,1).toUpperCase()+word.substring(1,word.length());
    else
    	wordList[wordPosition]=word;
    wordPosition++;
  }

  private String finishPanel(String ending) {
    theArea.setText(wordList[0]);
    for (int x=1; x<wordList.length; x++) {
      theArea.append("\n"+wordList[x]);
    }
    theArea.append(ending);
    return (new String(charList));
  }

  private void initWordsUsed() {
  	int x;
    for (x=0; x<adverbsUsed.length; x++)
    	adverbsUsed[x]=false;
    for (x=0; x<adjectivesUsed.length; x++)
    	adjectivesUsed[x]=false;
    for (x=0; x<articlesUsed.length; x++)
    	articlesUsed[x]=false;
    for (x=0; x<nounsUsed.length; x++)
    	nounsUsed[x]=false;
    for (x=0; x<verbsUsed.length; x++)
    	verbsUsed[x]=false;
  }

  private boolean isPlural(String word) {
  	if (word.equals("all") ||
    		word.equals("their") ||
      	word.equals("many") ||
      	word.equals("some"))
			return true;
    else
    	return false;
  }

  private String pickAWord(String [] words, boolean [] used) {
  	int z=random(words.length);		//Pick the nth unused word
    //Not necessarily the most efficient, but we won't be using a high percentage
    while (used[z])
    	z=random(words.length);
    used[z]=true;
    return words[z];
  }
}


